"""Admin classes for the rapid_prototyping app."""
# from django.contrib import admin

# from . import models


# class YourModelAdmin(admin.ModelAdmin):
#    list_display = ['some', 'fields', ]
#    search_fields = ['some', 'fieds', ]

# admin.site.register(models.YourModel, YourModelAdmin)



"""URLs for the rapid_prototyping app."""
from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^sprints/$',
        views.SprintListView.as_view(),
        name='rapid_prototyping_sprint_list'),
)



"""Templatetags for the rapid_prototyping app."""
from django import template
from django.conf import settings


register = template.Library()


@register.filter
def calculate_costs(minutes, hourly_rate):
    """
    Returns the costs for a task based on time and hourly rate.

    Usage::

        {{ item.time|calculate_costs:HOURLY_RATE }}

    """
    return round(float(minutes) / 60 * hourly_rate, 2)


@register.simple_tag
def get_difference(time, actual_time):
    """
    Returns the difference in percent for two given times.

    If it returns 1 it means that the task was finished as planned.
    If it returns n > 1 it means that the task took n% longer than planned
    If it returns n < 1 it means that the task took n% lesser than planned

    """
    return round(float(actual_time) / float(time), 2)


@register.assignment_tag
def get_hourly_rate(item_rate):
    """
    Returns the item rate if given, otherwise the default rate.

    Usage::

        {% get_hourly_rate item.rate as hourly_rate %}
        {{ hourly_rate %}

    """
    if item_rate is None or item_rate is '':
        return settings.RAPID_PROTOTYPING_HOURLY_RATE
    return item_rate



"""Tests for the templatetags of the rapid_prototyping app."""
from django.test import TestCase

from ..templatetags.rapid_prototyping_tags import (
    calculate_costs,
    get_difference,
    get_hourly_rate,
)


class CalculateCostTestCase(TestCase):
    """Tests for the ``calculate_costs`` template filter."""
    longMessage = True

    def test_tag(self):
        result = calculate_costs(380, 70)
        self.assertEqual(result, round(float(380) / 60 * 70, 2), msg=(
            'Tag should assume that the first parameter represents minutes'
            ' and the second parameter represents an hourly rate. It should'
            ' Correctly calculate the costs for the task based on the minutes'
            ' and the rate.'))


class GetDifferenceTestCase(TestCase):
    """Tests for the ``get_difference`` simple tag."""
    longMessage = True

    def test_tag(self):
        result = get_difference(100, 100)
        self.assertEqual(result, 1, msg=(
            'If there is no difference, it should return 1'))
        result = get_difference(100, 50)
        self.assertEqual(result, 0.5, msg=(
            'If actual time was less than estimated time, it should return'
            ' something less than 1'))
        result = get_difference(100, 150)
        self.assertEqual(result, 1.5, msg=(
            'If actual time was more than estimated time, it should return'
            ' something bigger than 1'))


class GetHourlyRateTestCase(TestCase):
    """Tests for the ``get_hourly_rate`` assignment tag."""
    longMessage = True

    def test_tag(self):
        result = get_hourly_rate(None)
        self.assertEqual(result, 70, msg=(
            'If not item rate has been given, the default hourly rate should'
            ' be returned'))

        result = get_hourly_rate(50)
        self.assertEqual(result, 50, msg=(
            'If an item rate is given, it should override the default hourly'
            ' rate and be returned.'))

        result = get_hourly_rate(0)
        self.assertEqual(result, 0, msg=(
            'If the item rate is given and it is 0, it should return 0. This'
            ' is to enable defining tasks that are given away for free.'))



"""Views for the rapid_prototyping app."""
from django.views.generic import TemplateView

from .context.utils import get_sprints


class SprintListView(TemplateView):
    template_name = 'rapid_prototyping/sprints.html'

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        return super(SprintListView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(SprintListView, self).get_context_data(**kwargs)
        sprints = get_sprints()
        ctx.update({
            'sprints': sprints,
        })
        return ctx



"""Tests for the views of the rapid_prototyping app."""
from mock import patch

from django.test import TestCase, RequestFactory

from ..views import SprintListView


class SprintListViewTestCase(TestCase):
    """Tests for the ``SprintListView`` class based view."""
    longMessage = True

    @patch('rapid_prototyping.views.get_sprints')
    def test_view(self, get_sprints_mock):
        req = RequestFactory().get('/')
        view = SprintListView()
        resp = view.dispatch(req)
        self.assertEqual(resp.status_code, 200, msg=(
            'View is callable'))
        self.assertTrue('sprints' in resp.context_data, msg=(
            'Sprints have been added to the context data'))



"""Utility functions that are useful for template context modules."""
from operator import itemgetter
import pkgutil

from django.conf import settings

from django_libs.loaders import load_member_from_setting, load_member


def get_sprints():
    """
    Returns all sprints, enriched with their assigned tasks.

    The project should only have one ``sprints.py`` module. We will define it's
    path via the ``RAPID_PROTOTYPING_SPRINTS_MODULE`` setting. The setting
    should be the fully qualified name of the ``sprints.py`` module (i.e.
    ``projectname.context.sprints.sprints``).

    Furthermore the project can have any amount of ``*_costs.py`` modules in
    any folder (as long as they are on the pythonpath).

    This function will find all ``*_costs.py`` modules and add those tasks,
    that have been assigned to a sprint, to the corresponding sprints in the
    ``sprints.py`` module.

    """
    sprints = load_member_from_setting(
        'RAPID_PROTOTYPING_SPRINTS_MODULE')

    all_tasks = []
    # TODO The onerror parameter is basically a workaround to ignore errors
    # The reason for that being, that in my case, the GeoDjango package was in
    # the path, permanently requesting certain libraries on import. Since they
    # were not present, the search was aborted with an OSError.
    for importer, package_name, _ in pkgutil.walk_packages(
            onerror=lambda p: p):
        if not package_name.endswith('_costs'):
            continue
        if not getattr(settings, 'TEST_RUN', None) and (
                '.test_app.' in package_name):
            continue
        costs = load_member(package_name + '.costs')
        for task in costs:
            all_tasks.append(task)
    sorted_tasks = sorted(all_tasks, key=itemgetter('id'))

    for sprint in sprints:
        remaining_time = 0
        sprint['tasks'] = []
        for task in sorted_tasks:
            if task.get('sprint') == sprint.get('id'):
                if not task.get('actual_time'):
                    remaining_time += \
                        task.get('developer_time') or task.get('time')
                sprint.get('tasks').append(task)
        sprint['remaining_time'] = remaining_time
        sprint['remaining_hours'] = round(float(remaining_time) / 60, 2)
    return sprints


def get_counter(counter):
    """
    Increases the counter by 1 and returns the changed list.

    This is a little hack. It works because lists are mutable. So in your
    ``context/home_costs.py`` you would something like this::

        from rapid_prototyping.context.utils import get_counter
        MAIN_ID = 100
        counter = [-1]
        costs = {
            'id': MAIN_ID + get_counter(counter)[0],
            ...
        }

    :param counter: A list which has one item, which is an integer, i.e.
      ``[-1]``.

    """
    counter[0] = counter[0] + 1
    return counter


def append_overhead_costs(costs, new_id, overhead_percentage=0.15):
    """
    Adds 15% overhead costs to the list of costs.

    Usage::

        from rapid_prototyping.context.utils import append_overhead_costs
        costs = [
            ....
        ]
        costs = append_overhead_costs(costs, MAIN_ID + get_counter(counter)[0])

    :param costs: Your final list of costs.
    :param new_id: The id that this new item should get.

    """
    total_time = 0
    for item in costs:
        total_time += item['time']

    costs.append({
        'id': new_id,
        'task': 'Overhead, Bufixes & Iterations',
        'time': total_time * overhead_percentage, },
    )
    return costs



from rapid_prototyping.context.utils import append_overhead_costs, get_counter


MAIN_ID = 100
counter = [-1]

costs = [
    {
        'id': MAIN_ID + get_counter(counter)[0],
        'task': 'First task',
        'time': 90,
        'developer_time': 110,
        'actual_time': 150,
        'link': 'http://www.google.com',
        'sprint': 1,
    },
    {
        'id': MAIN_ID + get_counter(counter)[0],
        'task': 'Second task',
        'time': 60,
    },
]

costs = append_overhead_costs(costs, MAIN_ID + get_counter(counter)[0])



"""Tests for the context utility functions."""
from django.test import TestCase

from ..context.utils import append_overhead_costs, get_counter, get_sprints


class AppendOverheadCostsTestCase(TestCase):
    """Tests for the ``append_overhead_costs`` function."""
    longMessage = True

    def test_function(self):
        costs = [
            {
                'id': 1,
                'task': 'Test',
                'time': 60,
            },
            {
                'id': 2,
                'task': 'Test2',
                'time': 60,
            },
        ]
        costs = append_overhead_costs(costs, 3)
        self.assertEqual(len(costs), 3, msg=(
            'The function should add a new item to the list.'))
        self.assertEqual(costs[2]['time'], 120 * 0.15, msg=(
            'The function should add 15% to the total costs.'))


class GetCounterTestCase(TestCase):
    """Tests for the ``get_counter`` function."""
    longMessage = True

    def test_function(self):
        counter = [-1]
        get_counter(counter)
        self.assertEqual(counter[0], 0, msg=(
            'The function should change the list and increase the value by'
            ' one.'))


class GetSprintsTestCase(TestCase):
    """Tests for the ``get_sprints`` function."""
    longMessage = True

    def test_function(self):
        result = get_sprints()
        self.assertEqual(result[0]['id'], 1, msg=(
            'Should return the first sprint as set up in the test_app'))
        self.assertEqual(result[0]['tasks'][0]['id'], 100, msg=(
            'Should add the tasks that belong to this sprint'))
